call = require('ramda/src/call'); contains = require('ramda/src/contains'); difference = require('ramda/src/difference'); has = require('ramda/src/has'); head = require('ramda/src/head'); identity = require('ramda/src/identity'); intersection = require('ramda/src/intersection'); isEmpty = require('ramda/src/isEmpty'); join = require('ramda/src/join'); keys = require('ramda/src/keys'); match = require('ramda/src/match'); merge = require('ramda/src/merge'); replace = require('ramda/src/replace'); test = require('ramda/src/test'); toLower = require('ramda/src/toLower'); toUpper = require('ramda/src/toUpper'); type = require('ramda/src/type'); union = require('ramda/src/union'); # auto_require: srcramda
{toPair, $, sf0} = RE = require 'ramda-extras' # auto_require: ramda-extras
[ːDateTime, ːDate, ːInt〳, ːStr, ːID_Int_Seq, ːBigInt, ːDateTime〳, ːBool〳, ːFloat〳, ːStr〳, ːBigInt〳, ːBool, ːFloat, ːInt, ːDate〳] = ['DateTime', 'Date', 'Int〳', 'Str', 'ID_Int_Seq', 'BigInt', 'DateTime〳', 'Bool〳', 'Float〳', 'Str〳', 'BigInt〳', 'Bool', 'Float', 'Int', 'Date〳'] #auto_sugar

{_extractWhere, fieldsClauseAndChildren} = query = require './query'
{AND, _, PVE, NYIE} = utils = require './utils'

_camelToSnake = (s) -> $ s, replace /[A-Z]/g, (s) -> '_' + toLower s
_snakeToCamel = (s) -> $ s, replace /_[a-z]/g, (s) -> toUpper s[1]
_pascalToSnake = (s) -> $ s, replace(/^./, toLower s[0]), replace /[A-Z]/g, (s) -> '_' + toLower s
_snakeToPascal = (s) -> $ s, replace(/^./, toUpper s[0]), replace /_[a-z]/g, (s) -> toUpper s[1]



createModel = (spec) ->
	model = {$config: {entityToTable: identity, tableToEntity: identity,
	propToCol: identity, colToProp: identity}} # default config

	forAll = (entity) -> {}
	# handle config
	if spec.$config
		for k, v of spec.$config
			if k == 'entityToTable'
				if type(v) == 'Function' then entityToTable = v
				else if v == 'pascalToSnake'
					model.$config.entityToTable = _pascalToSnake
					model.$config.tableToEntity = _snakeToPascal
				else throw new PVE "invalid config for entityToTable, '#{sf0 v}'"
			else if k == 'propToCol'
				if type(v) == 'Function' then propToCol = v
				else if v == 'camelToSnake'
					model.$config.propToCol = _camelToSnake
					model.$config.colToProp = _snakeToCamel
				else throw new PVE "invalid config for propToCol, '#{sf0 v}'"
			else if k == 'forAll'
				if type(v) != 'Function' then new PVE "config.forAll needs to be a function"
				model.$config.forAll = forAll = v
			else throw new PVE "invalid config property #{k}"

	# first pass for fields and temporary $rels and $subs
	for entity, fields of spec
		if entity == '$config' then continue

		if entity[0] != toUpper(entity[0])
			throw new PVE "entity must begin with uppercase character, see entity '#{entity}'"
		# o = model[entity] = {fields: {}}
		$fields = {}
		$rels = {}

		$subs = {}
		# $fns = {}

		# $unknownFields = []

		for k, v of fields
			switch v
				when ːStr, ːInt, ːBigInt, ːBool, ːFloat, ːDate, ːDateTime then $fields[k] = v
				when ːStr〳, ːInt〳, ːBigInt〳, ːBool〳, ːFloat〳, ːDate〳, ːDateTime〳 then $fields[k] = v
				when ːID_Int_Seq then $fields[k] = v
				else
					if type(v) == 'Function'
						# args = match(/\{(.*)\}/, v.toString())[1]
						# deps = split ', ', args
						$subs[k] = {fun: v}
					else
						if type(v) != 'Object'
							throw new PVE "unknown field type '#{v}' in model '#{entity}'"

						[k1, v1] = toPair v
						switch k1
							when 'oneToOne', 'oneToMany', 'manyToOne', 'oneToOne〳', 'oneToMany〳', 'manyToOne〳'
								if type(v1) != 'String' then throw new PVE "invalid relationship '#{v1}'"
								$rels[k] = {type: k1, link: v1}
							else
								$subs[k] = {obj: {k: k1, v: v1}}

		model[entity] = {fields: merge($fields, forAll(entity)), entity, $rels, $subs}

	# second pass for rels
	for entity, def of model
		if def.$rels
			def.rels = {}
			for k, v of def.$rels
				[___, e1, e1p, e2, e2p] = match /(.*?)\.(.*)\s=\s(.*?)\.(.*)/, v.link
				if ! contains e1, keys(model)
					throw new PVE "invalid link '#{v.link}' to non-existing entity '#{e1}'"
				if ! contains e2, keys(model)
					throw new PVE "invalid link '#{v.link}' to non-existing entity '#{e2}'"
				if e1 != entity
					throw new PVE "first entity '#{e1}' need to be '#{entity}' in link '#{v.link}'"
				if e1p == k
					throw new PVE """linking column '#{e1p}' cannot have same name as relationship \
					itself ('#{k}') for entity '#{entity}'"""

				fromForeignType = (dataType) ->
					if dataType == ːID_Int_Seq then ːInt
					else dataType

				if ! has e1p, def.fields
					def.fields[e1p] = fromForeignType model[e2].fields[e2p]
				else if ! has e2p, model[e2].fields then model[e2].fields[e2p] = fromForeignType def.fields[e1p]


				multiplicity = replace /〳$/g, '', v.type
				notNull = ! test /〳$/, v.type
				def.rels[k] = {multiplicity, notNull, entity: e2, on: [e1p, e2p], isRel: true}
				
			delete def.$rels


	# third pass for sub keys, return types, wheres, aggs
	for entity, def of model
		if def.$subs
			for k, $sub of def.$subs
				{obj, fun} = $sub
				if obj
					$sub.key = obj.k
					$sub.where = _extractWhere {}, obj.v[0]
					$sub.aggregates = query._extractAggregates {}, obj.v[0]
				else if fun
					try 
						[args] = utils.parseArguments $sub.fun.toString()
						res = fun args
						$sub.key = $ res, keys, head
						aggregates = query._extractAggregates {}, res[$sub.key][0]
						if !isEmpty aggregates then $sub.aggregates = aggregates
						fieldGroupBy = query._extractGroupBy res[$sub.key][0]
						if !isEmpty fieldGroupBy then $sub.groupBy = fieldGroupBy
						# [fieldsClause, children] = fieldsClauseAndChildren null, res
						# qq 2, [fieldsClause, children]
					catch err
						console.error err
						throw new PVE "sub #{k} in entity #{entity} threw an error on initial call, see docs"

	# fourth pass to resolve subs
	for entity, def of model
		if def.$subs
			def.subs = {}
			for k, $sub of def.$subs
				siblingDeps = []
				parentFieldDeps = []
				funArgs = {}
				if $sub.fun
					[funArgs] = utils.parseArguments $sub.fun.toString()

					# parent field deps
					parentFieldDeps = intersection keys(funArgs), keys def.fields

					# sibling deps
					siblingDeps = difference keys(funArgs), keys def.fields
					unknownDeps = difference siblingDeps, union keys(def.rels), keys(def.$subs)
					if !isEmpty unknownDeps
						throw new PVE "Your sub #{k}, has unknown dependencies #{join(', ', unknownDeps)}"
					if contains k, siblingDeps
						throw new PVE "Your sub #{k}, has a dependency on itself"


				composedWhere = $sub.where
				reachedFunction = false
				newQuery = null
				key = $sub.key
				lap = 0
				while !def.rels[key] && !newQuery
					parentSub = def.$subs[key]
					newQuery = model[key]
					if !parentSub && !newQuery
						throw new PVE "cannot resolve sub #{k} in model. #{sf0 $sub}. No entity or sibling #{key}"
					if reachedFunction then continue
					if newQuery then break

					if parentSub.obj
						composedWhere = utils.AND parentSub.obj.v[0], composedWhere
					else if parentSub.fun
						reachedFunction = true
						# siblingDeps.push key borde inte behövas längre

					key = parentSub.key
					if lap++ > 20 then throw new PVE "sub #{k} does not resolve. #{sf0 $sub}"

				subType = if $sub.obj then 'object' else 'function'
				if newQuery
					def.subs[k] = {entity: key, siblingDeps, funArgs, parentFieldDeps, isSub: true, subType,
					fun: $sub.fun, newQuery: true}
				else
					def.subs[k] = {entity: def.rels[key].entity, siblingDeps, funArgs,
					parentFieldDeps, isSub: true, subType, rel: key, fun: $sub.fun}
				if $sub.obj then def.subs[k].where = composedWhere
				if !isEmpty $sub.aggregates
					def.subs[k].aggregates = $sub.aggregates
					def.subs[k].isAggregate = true
					def.subs[k].returnType = 'aggregateValue'
				if $sub.groupBy
					def.subs[k].groupBy = $sub.groupBy
					def.subs[k].returnType = 'grouped'


			delete def.$subs

	# fifth pass to clarify siblingDeps
	# TODO: kanske ta bort denna och bara använda siblingDeps
	for entity, def of model
		if def.subs
			for k, sub of def.subs
				relDeps = []
				aggDeps = []
				groupDeps = []
				objDeps = []
				funDeps = []
				for sib in sub.siblingDeps
					sibRel = def.rels[sib]
					if sibRel then relDeps.push sib
					else
						sibSub = def.subs[sib]
						if sibSub.groupBy then groupDeps.push sib
						if sibSub.aggregates then aggDeps.push sib
						else if sibSub.subType == 'object' then objDeps.push sib
						else if sibSub.subType == 'function' then funDeps.push sib

				sub.siblingDepsGrouped = {relDeps, aggDeps, groupDeps, objDeps, funDeps}






	# qqq 9, model




	return model


module.exports = {
	createModel
	_snakeToCamel
	_camelToSnake
	_snakeToPascal
	_pascalToSnake
}





















